/*
 *	qnx4.c				Terminal routines for QNX V4.0
 *
 *	Created:	18/2/92		jmh
 *	Modified:	8/3/92		jmh
 *		New for March - mouse support!
 *
 *	The routines in this file provide screen and keyboard support
 *	using the QNX term_ library.  They therefore should work on
 *	the QNX console and also terminals.  It seems to work OK
 *	on a VT100 clone.
 *
 *	Uses the console in whatever video mode it finds it in.  No
 *	provision is made for switching VGA->EGA->CGA->whatever.
 *	Also, I'm not at all sure if bright backgrounds will work.
 *
 *	This module constructed using the os2 NPM module as a base.
 *
 *	Brought to you by Jim Hague, 1331 Ridgefield Road,
 *	Oxford OX4 3BZ, U.K. Tel: +44 (0)865 772688.
 *	`QNX4 drivers and data logging a speciality'
 *
 *************************************************************
 *  Slightly hacked for uemacs 3.12 / QNX 4.2 / Watcom 9.5
 *  by Trever Miller, December 20, 1993
 *  quics: tamiller
 *  inet : millertr@cuug.ab.ca  ||  trever.miller@cyberdex.cuug.ab.ca
 *
 *   + got the window-resize code to work.
 *
 *   + added posix termios code to allow ^C and ^S/^Q as input
 *
 *     NOTE: Sometimes the system still snags the ^C as interrupt.
 *           It appears to only happen with ^X^C in quick succession.
 *           Therefore, put about 1/2 or 1 sec inbetween the ^X and ^C.
 *
 *
 */

#define	termdef	1			/* don't define "term" external */
#include <env.h>
#include <conio.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/qnxterm.h>

#include "estruct.h"
#include "eproto.h"
#include "edef.h"
#include "elang.h"

#if QNX4				/* Only compile for QNX v4 */

/* The mouse on QNX4.01 screams around the screen at a hell of a rate.
 * If SLOW_MOUSE is 1, we check for environment variables
 * TERM_MOUSE_{X|Y}SKIP and if not found created them both with
 * values of {X|Y}_MOUSE_SKIP before switching the mouse on.
 */
#define	SLOW_MOUSE		1
#define	X_MOUSE_SKIP		"4"
#define	Y_MOUSE_SKIP		"4"

#define	EVENT_BUF_SIZE		10

#define	NROW    50		/* Screen size.                 */
#define	NCOL    80		/* Edit if you want to.         */

#define MAXNROW 100 /* however big a wterm window can get? */
#define MAXNCOL 200 
#define FORMFEED 12

#define	MARGIN	8		/* size of minimim margin and	*/
#define	SCRSIZ	64		/* scroll size for extended lines */
#define	NPAUSE	100		/* # times thru update to pause */

/* Forward references.          */

PASCAL NEAR qnx4move();
PASCAL NEAR qnx4eeol();
PASCAL NEAR qnx4eeop();
PASCAL NEAR qnx4clrdesk();
PASCAL NEAR qnx4beep();
PASCAL NEAR qnx4open();
PASCAL NEAR qnx4close();
PASCAL NEAR qnx4getc();
PASCAL NEAR qnx4putc();
PASCAL NEAR qnx4flush();
PASCAL NEAR qnx4rev();
PASCAL NEAR qnx4kclose();
PASCAL NEAR qnx4kopen();
PASCAL NEAR qnx4cres();
PASCAL NEAR qnx4parm();
#if	COLOR
PASCAL NEAR qnx4fcol();
PASCAL NEAR qnx4bcol();
#endif

static struct termios t_attr, orig_t_attr;

static int	mouse = 0;		/* Have we got a mouse ? */

static int	cfcolour = 7;		/* current foreground colour */
static int	cbcolour = 0;		/* current background colour */

static unsigned	ctrans_fg[] = /* ansi to qnx foreground translation table */
	{ TERM_BLACK, TERM_RED, TERM_GREEN, TERM_YELLOW, TERM_BLUE,
	  TERM_MAGENTA, TERM_CYAN, TERM_WHITE,
	  TERM_BLACK | TERM_HILIGHT, TERM_RED | TERM_HILIGHT,
	  TERM_GREEN | TERM_HILIGHT, TERM_YELLOW | TERM_HILIGHT,
	  TERM_BLUE | TERM_HILIGHT, TERM_MAGENTA | TERM_HILIGHT,
	  TERM_CYAN | TERM_HILIGHT, TERM_WHITE | TERM_HILIGHT };
	  
static unsigned	ctrans_bg[] = /* ansi to qnx background translation table */
	{ TERM_BLACK_BG, TERM_RED_BG, TERM_GREEN_BG, TERM_YELLOW_BG,
	  TERM_BLUE_BG, TERM_MAGENTA_BG, TERM_CYAN_BG, TERM_WHITE_BG,
	  TERM_BLACK_BG, TERM_RED_BG, TERM_GREEN_BG, TERM_YELLOW_BG,
	  TERM_BLUE_BG, TERM_MAGENTA_BG, TERM_CYAN_BG, TERM_WHITE_BG };

static short	 qnx4row;		/* current cursor row	*/
static short	 qnx4col;		/* current cursor col	*/

static int revflag = FALSE;		/* are we currently in rev video? */

/* Sometime term_fill will set the foreground characters to " rather
 * than space.  I suspect this is something going wrong with
 * HILIGHT attributes, so mask 'em off for now.
 */
#define	FILL_MASK	(~TERM_HILIGHT) 

/*
 * Standard terminal interface dispatch table.
 */
TERM    term    = {
	MAXNROW - 1,
	NROW - 1,
	MAXNCOL,
	NCOL,
	0, 0,
	MARGIN,
	SCRSIZ,
	NPAUSE,
	qnx4open,
	qnx4close,
	qnx4kopen,
	qnx4kclose,
	qnx4getc,
	qnx4putc,
	qnx4flush,
	qnx4move,
	qnx4eeol,
	qnx4eeop,
	qnx4clrdesk,
	qnx4beep,
	qnx4rev,
	qnx4cres
#if	COLOR
	, qnx4fcol,
	qnx4bcol
#endif
};

static struct {
	unsigned qnx_key;		/* QNX term_ keypress code */
	unsigned emacs_key;		/* uEmacs key code */
} keytable[] = {
	K_F1,	SPEC | '1',		/* Function keys - normal */
	K_F2,	SPEC | '2',
	K_F3,	SPEC | '3',
	K_F4,	SPEC | '4',
	K_F5,	SPEC | '5',
	K_F6,	SPEC | '6',
	K_F7,	SPEC | '7',
	K_F8,	SPEC | '8',
	K_F9,	SPEC | '9',
	K_F10,	SPEC | '0',
	K_F11,	SPEC | '-',
	K_F12,	SPEC | '=',
	
	K_SHF_F1,	SPEC | SHFT | '1',	/* Function keys - shifted */
	K_SHF_F2,	SPEC | SHFT | '2',
	K_SHF_F3,	SPEC | SHFT | '3',
	K_SHF_F4,	SPEC | SHFT | '4',
	K_SHF_F5,	SPEC | SHFT | '5',
	K_SHF_F6,	SPEC | SHFT | '6',
	K_SHF_F7,	SPEC | SHFT | '7',
	K_SHF_F8,	SPEC | SHFT | '8',
	K_SHF_F9,	SPEC | SHFT | '9',
	K_SHF_F10,	SPEC | SHFT | '0',
	K_SHF_F11,	SPEC | SHFT | '-',
	K_SHF_F12,	SPEC | SHFT | '=',

	K_CTL_F1,	SPEC | CTRL | '1',	/* Function keys - control */
	K_CTL_F2,	SPEC | CTRL | '2',
	K_CTL_F3,	SPEC | CTRL | '3',
	K_CTL_F4,	SPEC | CTRL | '4',
	K_CTL_F5,	SPEC | CTRL | '5',
	K_CTL_F6,	SPEC | CTRL | '6',
	K_CTL_F7,	SPEC | CTRL | '7',
	K_CTL_F8,	SPEC | CTRL | '8',
	K_CTL_F9,	SPEC | CTRL | '9',
	K_CTL_F10,	SPEC | CTRL | '0',
	K_CTL_F11,	SPEC | CTRL | '-',
	K_CTL_F12,	SPEC | CTRL | '=',

	K_ALT_F1,	SPEC | ALTD | '1',	/* Function keys - alt */
	K_ALT_F2,	SPEC | ALTD | '2',
	K_ALT_F3,	SPEC | ALTD | '3',
	K_ALT_F4,	SPEC | ALTD | '4',
	K_ALT_F5,	SPEC | ALTD | '5',
	K_ALT_F6,	SPEC | ALTD | '6',
	K_ALT_F7,	SPEC | ALTD | '7',
	K_ALT_F8,	SPEC | ALTD | '8',
	K_ALT_F9,	SPEC | ALTD | '9',
	K_ALT_F10,	SPEC | ALTD | '0',
	K_ALT_F11,	SPEC | ALTD | '-',
	K_ALT_F12,	SPEC | ALTD | '=',

	K_ALT_A,	ALTD | 'A',		/* Alt letters */
	K_ALT_B,	ALTD | 'B',
	K_ALT_C,	ALTD | 'C',
	K_ALT_D,	ALTD | 'D',
	K_ALT_E,	ALTD | 'E',
	K_ALT_F,	ALTD | 'F',
	K_ALT_G,	ALTD | 'G',
	K_ALT_H,	ALTD | 'H',
	K_ALT_I,	ALTD | 'I',
	K_ALT_J,	ALTD | 'J',
	K_ALT_K,	ALTD | 'K',
	K_ALT_L,	ALTD | 'L',
	K_ALT_M,	ALTD | 'M',
	K_ALT_N,	ALTD | 'N',
	K_ALT_O,	ALTD | 'O',
	K_ALT_P,	ALTD | 'P',
	K_ALT_Q,	ALTD | 'Q',
	K_ALT_R,	ALTD | 'R',
	K_ALT_S,	ALTD | 'S',
	K_ALT_T,	ALTD | 'T',
	K_ALT_U,	ALTD | 'U',
	K_ALT_V,	ALTD | 'V',
	K_ALT_W,	ALTD | 'W',
	K_ALT_X,	ALTD | 'X',
	K_ALT_Y,	ALTD | 'Y',
	K_ALT_Z,	ALTD | 'Z',

	K_HOME,		SPEC | '<',		/* Keypad */
	K_UP,		SPEC | 'P',
	K_PGUP,		SPEC | 'Z',
	K_KPDMINUS,	'-',
	K_LEFT,		SPEC | 'B',
	K_KPDFIVE,	SPEC | 'L',
	K_RIGHT,	SPEC | 'F',
	K_KPDPLUS,	'+',
	K_END,		SPEC | '>',
	K_DOWN,		SPEC | 'N',
	K_PGDN,		SPEC | 'V',
	K_INSERT,	SPEC | 'C',
	K_DELETE,	SPEC | 'D',
	
	K_CTL_HOME,	SPEC | CTRL | '<',	/* Keypad control */
	K_CTL_UP,	SPEC | CTRL | 'P',
	K_CTL_PGUP,	SPEC | CTRL | 'Z',
	K_CTL_KPDMINUS,	'-',
	K_CTL_LEFT,	SPEC | CTRL | 'B',
	K_CTL_KPDFIVE,	SPEC | CTRL | 'L',
	K_CTL_RIGHT,	SPEC | CTRL | 'F',
	K_CTL_KPDPLUS,	'+',
	K_CTL_END,	SPEC | CTRL | '>',
	K_CTL_DOWN,	SPEC | CTRL | 'N',
	K_CTL_PGDN,	SPEC | CTRL | 'V',
	K_CTL_INSERT,	SPEC | CTRL | 'C',
	K_CTL_DELETE,	SPEC | CTRL | 'D',
	
	K_BACKTAB,	SHFT | CTRL | 'I',	/* Misc */
	0,		0			/* End of table */
};

/************************* Local functions ****************************/

/* Originally a macro, this returns the screen attribute corresponding
 * to the given colours.  As a special gross hack for mono screens,
 * we look out for white on black with reverse requested, and do
 * this as white on black plus inverse.  This is so they get an
 * inverse mode line. */
static unsigned ATTR(unsigned fg, unsigned bg)
{
	unsigned res;

	if ( fg == 7 && bg == 0 && revflag )
		res = TERM_WHITE | TERM_BLACK_BG | TERM_INVERSE;
	else
		res = ((revflag) ? (ctrans_fg[bg] | ctrans_bg[fg])
				 : (ctrans_fg[fg] | ctrans_bg[bg]));

	return (res);

}

static unsigned char event_buf[EVENT_BUF_SIZE];
static int ev_buf_in = 0, ev_buf_out = 0;

static void ev_init(void)
{
	ev_buf_in = ev_buf_out = 0;
}

static int ev_check(void)
{
	return (ev_buf_in != ev_buf_out);
}

static unsigned char ev_get(void)
{
	unsigned char res = 0;
	
	if ( ev_buf_in != ev_buf_out )
	{
		res = event_buf[ev_buf_out++];
		if ( ev_buf_out >= EVENT_BUF_SIZE )
			ev_buf_out = 0;
	}
	return (res);
}

static void ev_put(unsigned char event)
{
	event_buf[ev_buf_in++] = event;
	if ( ev_buf_in >= EVENT_BUF_SIZE )
		ev_buf_in = 0;
}

/************************* Public functions ****************************/

#if	COLOR
/*----------------------------------------------------------------------*/
/*	qnx4fcol()							*/
/* Set the current foreground colour.					*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4fcol(int colour)
{
	cfcolour = colour;
	term_color (ctrans_fg[cfcolour] | ctrans_bg[cbcolour]);
}

/*----------------------------------------------------------------------*/
/*	qnx4bcol()							*/
/* Set the current background colour.					*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4bcol(int colour)
{
	cbcolour = colour;
	
	term_fill (ctrans_bg[cbcolour]);
	term_color (ctrans_fg[cfcolour] | ctrans_bg[cbcolour]);
}
#endif


/*----------------------------------------------------------------------*/
/*	qnx4move()							*/
/* Move the cursor.							*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4move(int row,int col)
{
	qnx4row = row;
	qnx4col = col;
	term_cur (row, col);
}


/*----------------------------------------------------------------------*/
/*	qnx4flush()							*/
/* Update the physical video buffer from the logical video buffer.	*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4flush(void)
{
	term_flush ();
}


/*----------------------------------------------------------------------*/
/*	qnx4getc()							*/
/* Get a character from the keyboard.					*/
/* Munge function, edit, alt keys etc into the 0 followed by two	*/
/* bytes of extended code expected by MicroEmacs.			*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4getc(void)
{
	unsigned res;
	unsigned qnx_key;
	unsigned char m_event;
	int i;

	/* Any events waiting ? */
	if ( ev_check () )
		return (ev_get ());
	
	/* Get a qnx keypress */
	while((res = qnx_key = term_key()) == K_RESIZE)
	{
		newwidth(TRUE,term_state.num_cols);
		newsize(TRUE,term_state.num_rows - 1);
		update(TRUE);
	}

#if	MOUSE
	/* Is it a mouse event ? */
	if ( mouse && (qnx_key & K_CLASS) == K_MOUSE_POS )
	{
		ev_put (MOUS >> 8);
		ev_put (term_state.mouse_col);
		ev_put (term_state.mouse_row);

		if ( (qnx_key & K_MOUSE_ACTION) == K_MOUSE_RELEASE )
		{
			m_event = 1;
			term_mouse_hide ();
		}
		else
			m_event = 0;

		if ( (qnx_key & K_MOUSE_BUTTONS) == K_MOUSE_BRIGHT )
			m_event += 4;
		if ( (qnx_key & K_MOUSE_BUTTONS) == K_MOUSE_BMIDDLE )
			m_event += 2;

		/*
		 * Should test shift and control key state here,
		 * but I'm not sure if it's possible.
		 */
		m_event += 'a';

		ev_put (m_event);
		return (0);
	}
#endif

	/* Translate it if needs be */
	if ( qnx_key > 0xff )
	{
		res = qnx_key & 0xff;
		for ( i = 0 ; keytable[i].qnx_key != 0 ; i++ )
			if ( keytable[i].qnx_key == qnx_key )
			{
				res = keytable[i].emacs_key;
				break;
			}
	}

	if ( res > 0xff )		/* Got translated key */
	{
		ev_put((res >> 8) & 0xff);
		ev_put(res & 0xff);
		return (0);
	}
		
	return(res);
}

#if	TYPEAH
/*----------------------------------------------------------------------*/
/*	typahead()							*/
/* Returns true if a key has been pressed. Not sure if I've implemented	*/
/* this properly, or indeed if it matters.				*/
/*----------------------------------------------------------------------*/

PASCAL NEAR typahead(void)

{
	return kbhit();
}
#endif


/*----------------------------------------------------------------------*/
/*	qnx4putc()							*/
/* Put a character at the current position in the current colours.	*/
/* This routine does nothing with returns and linefeeds.  For backspace */
/* it puts a space in the previous column and moves the cursor to the	*/
/* previous column.  For all other characters, it will display the	*/
/* graphic representation of the character and put the cursor in the	*/
/* next column (even if that is off the screen.  In practice this isn't	*/
/* a problem).								*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4putc(int c)
{
	char ch = c;
	unsigned attr;
	char space[] = " ";
	
	if (c == '\n' || c == '\r') {		/* returns and linefeeds */
		return;
	}

	attr = ATTR(cfcolour, cbcolour);
	
	if (c == '\b' && qnx4col > 0) {			/* backspace		*/
		term_type (qnx4row, --qnx4col, &space, 1, attr);
		term_cur (qnx4row, qnx4col);
	}
	else {
		term_type (qnx4row, qnx4col, &ch, 1, attr);
		++qnx4col;			/* move cursor forward	*/
	}
}


/*----------------------------------------------------------------------*/
/*	qnx4eeol()							*/
/* Erase to end of line.						*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4eeol(void)
{
	term_fill (ATTR(cfcolour, cbcolour) & FILL_MASK);
	term_clear (TERM_CLS_EOL);
}


/*----------------------------------------------------------------------*/
/*	qnx4eeop()							*/
/* Erase to end of page.						*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4eeop(void)
{
	term_fill (ATTR(cfcolour, cbcolour) & FILL_MASK);
	term_clear (TERM_CLS_EOS);
}

/*----------------------------------------------------------------------*/
/*	qnx4clrdesk()							*/
/* Erase all of page.							*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4clrdesk(void)
{
	term_fill (ATTR(cfcolour, cbcolour) & FILL_MASK);
	term_clear (TERM_CLS_SCR); 
}

/*----------------------------------------------------------------------*/
/*	qnx4rev()							*/
/* Change reverse video state.  					*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4rev(int state)	/* TRUE = reverse, FALSE = normal */
{
	revflag = state;
}

/*----------------------------------------------------------------------*/
/*	qnx4cres()							*/
/* Change the screen resolution.					*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4cres(char *res)		/* name of desired video mode	*/
{
	return TRUE;
}


/*----------------------------------------------------------------------*/
/*	spal()								*/
/* Change pallette settings.  (Does nothing.)				*/
/*----------------------------------------------------------------------*/

PASCAL NEAR spal(char *dummy)
{
}


/*----------------------------------------------------------------------*/
/*	qnx4beep()							*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4beep(void)
{
	write (1, "\007", 1);
}


/*----------------------------------------------------------------------*/
/*	qnx4open()							*/
/* Open the terminal for business. When used to reopen the terminal,	*/
/* the terminal library crashes on a subsequent clear unless you set	*/
/* the cursor position beforehand, so to be on the safe side we set	*/
/* the cursor position to the bottom LHS of the screen.			*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4open(void)
{
	static int term_loaded = 0;

	if ( term_loaded )
	{
		t_attr = orig_t_attr;
		t_attr.c_iflag &= ~(IXON); /* No inbound XON/XOFF == allow ^S/^Q */
		t_attr.c_lflag &= ~(ISIG); /* No signal characters == allow ^C */
		t_attr.c_lflag &= ~(NOFLSH); 
		tcsetattr(STDIN_FILENO,TCSAFLUSH,&t_attr);

		term_init ();
		revflag = FALSE;
		qnx4clrdesk();
	}
	else
	{
#if SLOW_MOUSE
		if ( getenv ("TERM_MOUSE_XSKIP") == NULL &&
		     getenv ("TERM_MOUSE_YSKIP") == NULL )
		{
			setenv ("TERM_MOUSE_XSKIP", X_MOUSE_SKIP, 0);
			setenv ("TERM_MOUSE_YSKIP", Y_MOUSE_SKIP, 0);
		}
#endif
		tcgetattr(STDIN_FILENO,&orig_t_attr);
		t_attr = orig_t_attr;
		t_attr.c_iflag &= ~(IXON); /* No inbound XON/XOFF == allow ^S/^Q */
		t_attr.c_lflag &= ~(ISIG); /* No signal characters == allow ^C */
		t_attr.c_lflag &= ~(NOFLSH); 
		tcsetattr(STDIN_FILENO,TCSAFLUSH,&t_attr);

		if ( term_load () == -1 )
		{
			fprintf (stderr, "Unknown or inadequate terminal.\n");
			exit(1);
		}

		term_resize_on ();

		term.t_nrow = term_state.num_rows - 1;
		term.t_ncol = term_state.num_cols;

		revexist = TRUE;
		revflag = FALSE;

		term_loaded = 1;
	}

	term_fill (ctrans_bg[cbcolour]);
	qnx4move (term.t_nrow, 0);

	return 0;		/* Success */
}


/*----------------------------------------------------------------------*/
/*	qnx4close()							*/
/* Restore the original video settings.					*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4close(void)
{
	term_fill (TERM_WHITE | TERM_BLACK_BG);
	term_clear (TERM_CLS_SCRH);
	term_restore ();
	tcsetattr(STDIN_FILENO,TCSAFLUSH,&orig_t_attr);
}

/*----------------------------------------------------------------------*/
/*	qnx4kopen()							*/
/* Open the keyboard.							*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4kopen(void)
{
#if	MOUSE
	unsigned ch;	
#endif

	/* Mostly done by term_open in qnx4open */
	ev_init();

#if	MOUSE
	mouse = (term_mouse_on() == 0);
	ch = TERM_MOUSE_SELECT | TERM_MOUSE_MENU | TERM_MOUSE_RELEASE;
	term_mouse_flags(ch, ch);
#endif
}


/*----------------------------------------------------------------------*/
/*	qnx4kclose()							*/
/* Close the keyboard.							*/
/*----------------------------------------------------------------------*/

PASCAL NEAR qnx4kclose(void)
{
	/* Mostly done by term_restore in qnx4close */

#if	MOUSE
	if ( mouse )
		term_mouse_off ();
#endif
}

#if	FLABEL
PASCAL NEAR fnclabel(f, n)	/* label a function key */

int f,n;	/* default flag, numeric argument [unused] */

{
	/* on machines with no function keys...don't bother */
	return(TRUE);
}
#endif

#endif

